#include "precomp.h"

#define HOLES			8
#define SPAWNPERFRAME	750
#define PARTICLES		500000

static bool started = false;
static int frame = 0, delay = 1;
static float r = 2, fps = 0;
static timer fpstimer;

BlackHole::BlackHole( float px, float py, float pg )
{
	ox = px * SCRWIDTH, oy = py * SCRHEIGHT, g = pg;
}

void Game::UpdateBlackHoles()
{
	float hw = SCRWIDTH / 2, hh = SCRHEIGHT / 2;
	for (unsigned int i = 0; i < HOLES; i++)
	{
		float hcx = m_Hole[i]->ox - hw, hcy = m_Hole[i]->oy - hh;
		m_Hole[i]->x = hcx * sin( r ) + hcy * cos( r ) + hw;
		m_Hole[i]->y = hcx * cos( r ) - hcy * sin( r ) + hh;
		r += 0.0001f;
		if (r > ( PI * 2 )) r -= PI * 2;
	}
}

void Game::BuildBackdrop()
{
	__m256 bhx8 = _mm256_set_ps( m_Hole[0]->x, m_Hole[1]->x, m_Hole[2]->x, m_Hole[3]->x, m_Hole[4]->x, m_Hole[5]->x, m_Hole[6]->x, m_Hole[7]->x );
	__m256 bhy8 = _mm256_set_ps( m_Hole[0]->y, m_Hole[1]->y, m_Hole[2]->y, m_Hole[3]->y, m_Hole[4]->y, m_Hole[5]->y, m_Hole[6]->y, m_Hole[7]->y );
	__m256 bhg8 = _mm256_set_ps( m_Hole[0]->g, m_Hole[1]->g, m_Hole[2]->g, m_Hole[3]->g, m_Hole[4]->g, m_Hole[5]->g, m_Hole[6]->g, m_Hole[7]->g );
	bhg8 = _mm256_mul_ps( bhg8, _mm256_set1_ps( 250.0f ) );
	#pragma omp parallel for 
	for (int y = 0; y < SCRHEIGHT; y++)
	{
		Pixel* dst = screen->GetBuffer() + y * SCRWIDTH;
		float fy = (float)y;
		float fx = 0;
		__m256 fy8 = _mm256_set1_ps( fy );
		for (unsigned int x = 0; x < SCRWIDTH; x++, fx++)
		{
			__m256 g8 = _mm256_setzero_ps();
			__m256 dx8 = _mm256_sub_ps( bhx8, _mm256_set1_ps( fx ) );
			__m256 dy8 = _mm256_sub_ps( bhy8, fy8 );
			__m256 sd8 = _mm256_add_ps( _mm256_mul_ps( dx8, dx8 ), _mm256_mul_ps( dy8, dy8 ) );
			g8 = _mm256_add_ps( g8, _mm256_mul_ps( bhg8, _mm256_rcp_ps( sd8 ) ) );
			float g = g8.m256_f32[0] + g8.m256_f32[1] + g8.m256_f32[2] + g8.m256_f32[3] +
				g8.m256_f32[4] + g8.m256_f32[5] + g8.m256_f32[6] + g8.m256_f32[7];
			if (g > 1) g = 0;
			*dst++ = (int)(g * 255.0f);
		}
		dst += screen->GetPitch() - screen->GetWidth();
	}
}

static union { float px[PARTICLES]; __m256 px8[PARTICLES / 4]; };
static union { float py[PARTICLES]; __m256 py8[PARTICLES / 4]; };
static union { float pvx[PARTICLES]; __m256 pvx8[PARTICLES / 4]; };
static union { float pvy[PARTICLES]; __m256 pvy8[PARTICLES / 4]; };
static union { float pm[PARTICLES]; __m256 pm8[PARTICLES / 4]; };
static bool pa[PARTICLES];
static union { uint pc[PARTICLES]; __m256i pc8[PARTICLES / 4]; };

void Game::Init()
{
	m_Particle = new Particle*[PARTICLES];
	m_Hole = new BlackHole*[HOLES];
	m_Hole[0] = new BlackHole( 0.2f, 0.3f, 1.0f );
	m_Hole[1] = new BlackHole( 0.5f, 0.5f, 3.0f );
	m_Hole[2] = new BlackHole( 0.7f, 0.6f, 0.5f );
	m_Hole[3] = new BlackHole( 0.4f, 0.9f, 1.8f );
	m_Hole[4] = new BlackHole( 0.25f, 0.8f, 0.8f );
	m_Hole[5] = new BlackHole( 0.9f, 0.3f, 0.5f );
	m_Hole[6] = new BlackHole( 0.7f, 0.4f, 0.8f );
	m_Hole[7] = new BlackHole( 0.8f, 0.1f, 0.9f );
	for (unsigned int i = 0; i < PARTICLES; i++)
	{
		m_Particle[i] = new Particle();
		if (i & 1)
		{
			m_Particle[i]->m = 0.7f;
			m_Particle[i]->c = (255 << 16) + (255 << 8);
		}
		else
		{
			m_Particle[i]->m = 1.5f;
			m_Particle[i]->c = (255 << 16) + 50 + (50 << 8);
		}
	}
	fpstimer.reset();
	// convert to SoA
	for (int i = 0; i < PARTICLES; i++)
	{
		px[i] = m_Particle[i]->x;
		py[i] = m_Particle[i]->y;
		pvx[i] = m_Particle[i]->vx;
		pvy[i] = m_Particle[i]->vy;
		pa[i] = 0;
		pc[i] = m_Particle[i]->c;
		pm[i] = m_Particle[i]->m;
	}
}

void Game::SpawnParticle( int n )
{
	pa[n] = 1;
	px[n] = 0;
	py[n] = SCRHEIGHT * (0.8f + Rand( 0.02f ));
	pvx[n] = 0.95f + Rand( 0.1f );
	pvy[n] = -0.05f + Rand( 0.1f );
}

void Game::UpdateParticles()
{
	__m256 bhx8 = _mm256_set_ps( m_Hole[0]->x, m_Hole[1]->x, m_Hole[2]->x, m_Hole[3]->x, m_Hole[4]->x, m_Hole[5]->x, m_Hole[6]->x, m_Hole[7]->x );
	__m256 bhy8 = _mm256_set_ps( m_Hole[0]->y, m_Hole[1]->y, m_Hole[2]->y, m_Hole[3]->y, m_Hole[4]->y, m_Hole[5]->y, m_Hole[6]->y, m_Hole[7]->y );
	__m256 bhg8 = _mm256_set_ps( m_Hole[0]->g, m_Hole[1]->g, m_Hole[2]->g, m_Hole[3]->g, m_Hole[4]->g, m_Hole[5]->g, m_Hole[6]->g, m_Hole[7]->g );
	bhg8 = _mm256_mul_ps( bhg8, _mm256_set1_ps( 250.0f ) );
	for (unsigned int i = 0; i < PARTICLES; i++) if (pa[i])
	{
		px[i] += pvx[i];
		py[i] += pvy[i];
		if (!((px[i] < (2 * SCRWIDTH)) && (px[i] > -SCRWIDTH) &&
			(py[i] < (2 * SCRHEIGHT)) && (py[i] > -SCRHEIGHT)))
		{
			SpawnParticle( i );
			continue;
		}
	}
	for (unsigned int i = 0; i < PARTICLES; i++)
	{
		__m256 dx8 = _mm256_sub_ps( bhx8, _mm256_set1_ps( px[i] ) );
		__m256 dy8 = _mm256_sub_ps( bhy8, _mm256_set1_ps( py[i] ) );
		__m256 sd8 = _mm256_add_ps( _mm256_mul_ps( dx8, dx8 ), _mm256_mul_ps( dy8, dy8 ) );
		__m256 dist8 = _mm256_rsqrt_ps( sd8 );
		dx8 = _mm256_mul_ps( dx8, dist8 ), dy8 = _mm256_mul_ps( dy8, dist8 );
		__m256 g8 = _mm256_mul_ps( _mm256_set1_ps( 0.5f ), _mm256_div_ps( _mm256_mul_ps( bhg8, _mm256_set1_ps( pm[i] ) ), sd8 ) );
		__m256 gdx8 = _mm256_mul_ps( g8, dx8 );
		__m256 gdy8 = _mm256_mul_ps( g8, dy8 );
		for( int j = 0; j < 8; j++ ) 
			pvx[i] += gdx8.m256_f32[j],
			pvy[i] += gdy8.m256_f32[j];
		__m256 mask = _mm256_cmp_ps( g8, _mm256_set1_ps( 1.0f ), _CMP_GE_OS );
		if (_mm256_movemask_ps( mask ) > 0)
		{
			// particle fell into one of the black holes, don't care which one
			SpawnParticle( i );
		}
	}
	for (unsigned int i = 0; i < PARTICLES; i++) if (pa[i])
	{
		int x = (int)px[i], y = (int)py[i];
		if ((x >= 0) && (x < SCRWIDTH) && (y >= 0) && (y < SCRHEIGHT))
		{
			Pixel* dst = screen->GetBuffer();
			dst[x + y * screen->GetPitch()] = pc[i];
		}
	}
}

void Game::Tick( float dt )
{
	// spawn particles
	if (started) for (unsigned int i = 0; i < SPAWNPERFRAME; i++)
	{
		if (frame < PARTICLES)
		{
			SpawnParticle( frame );
			m_Particle[frame]->alive = true;
		}
		frame++;
	}
	if (GetAsyncKeyState( 32 )) started = true;

	// get accurate frame rate
	if (!--delay)
	{
		delay = 8;
		fps = 8000.0f / fpstimer.elapsed();
		fpstimer.reset();
	}

	// spin black holes
	timer t1;
	UpdateBlackHoles();
	float elapsed1 = t1.elapsed();

	// draw the gravity field
	timer t2;
	for (int i = 0; i < 10; i++) BuildBackdrop();
	float elapsed2 = t2.elapsed();

	// draw the particles	
	timer t3;
	UpdateParticles();
	float elapsed3 = t3.elapsed();

	// statistics
	char t[128];
	sprintf( t, "updateblackholes %5.1f ms", elapsed1 );
	screen->Print( t, 2, 2, 0xffffff );
	sprintf( t, "buildbackdrop    %5.1f ms", elapsed2 );
	screen->Print( t, 2, 12, 0xffffff );
	sprintf( t, "updateparticles  %5.1f ms", elapsed3 );
	screen->Print( t, 2, 22, 0xffffff );
	sprintf( t, "%i particles", min( PARTICLES, frame ) );
	screen->Print( t, 2, 32, 0xffffff );
	sprintf( t, "fps %6.2f", fps );
	screen->Print( t, 2, 42, 0xff9999 );
}